package eu.jonahbauer.raytracing.math;

import org.jetbrains.annotations.NotNull;

import java.util.Optional;
import java.util.random.RandomGenerator;

import static eu.jonahbauer.raytracing.Main.DEBUG;

public record Vec3(double x, double y, double z) {
    public static final Vec3 ZERO = new Vec3(0, 0, 0);
    public static final Vec3 MAX = new Vec3(Double.MAX_VALUE, Double.MAX_VALUE, Double.MAX_VALUE);
    public static final Vec3 MIN = new Vec3(-Double.MAX_VALUE, -Double.MAX_VALUE, -Double.MAX_VALUE);
    public static final Vec3 UNIT_X = new Vec3(1, 0, 0);
    public static final Vec3 UNIT_Y = new Vec3(0, 1, 0);
    public static final Vec3 UNIT_Z = new Vec3(0, 0, 1);

    public Vec3 {
        if (DEBUG) {
            if (!Double.isFinite(x) || !Double.isFinite(y) || !Double.isFinite(z)) {
                throw new IllegalArgumentException("x, y and z must be finite");
            }
        }
    }

    /**
     * {@return a uniformly random unit vector}
     */
    public static @NotNull Vec3 random(@NotNull RandomGenerator random) {
        double x, y, z;
        double squared;
        do {
            x = Math.fma(2, random.nextDouble(), -1);
            y = Math.fma(2, random.nextDouble(), -1);
            z = Math.fma(2, random.nextDouble(), -1);
            squared = x * x + y * y + z * z;
        } while (squared > 1);
        var factor = 1 / Math.sqrt(squared);
        return new Vec3(x * factor, y * factor, z * factor);
    }

    /**
     * {@return a uniformly random unit vector on the opposite hemisphere of the given <code>direction</code>}
     */
    public static @NotNull Vec3 randomOppositeHemisphere(@NotNull RandomGenerator random, @NotNull Vec3 direction) {
        double x, y, z;
        double squared;
        do {
            x = Math.fma(2, random.nextDouble(), -1);
            y = Math.fma(2, random.nextDouble(), -1);
            z = Math.fma(2, random.nextDouble(), -1);
            squared = x * x + y * y + z * z;
        } while (squared > 1 || direction.x() * x + direction.y() * y + direction.z() * z >= 0);
        var factor = 1 / Math.sqrt(squared);
        return new Vec3(x * factor, y * factor, z * factor);
    }

    /**
     * Reflects a vector on the given {@code normal} vector.
     * @param vec a vector
     * @param normal the surface normal (must be a unit vector)
     * @return the reflected vector
     */
    public static @NotNull Vec3 reflect(@NotNull Vec3 vec, @NotNull Vec3 normal) {
        var factor = - 2 * normal.times(vec);
        return Vec3.fma(factor, normal, vec);
    }

    /**
     * Refracts a vector on the given {@code normal} vector.
     * @param vec a vector
     * @param normal the surface normal (must be a unit vector)
     * @param ri the refractive index
     * @return the refracted vector
     */
    public static @NotNull Optional<Vec3> refract(@NotNull Vec3 vec, @NotNull Vec3 normal, double ri) {
        vec = vec.unit();
        var cosTheta = Math.min(- vec.times(normal), 1.0);
        var sinTheta = Math.sqrt(1 - cosTheta * cosTheta);
        if (ri * sinTheta > 1) return Optional.empty();

        var rOutPerp = vec.plus(normal.times(cosTheta)).times(ri);
        var rOutParallel = normal.times(- Math.sqrt(Math.abs(1 - rOutPerp.squared())));
        return Optional.of(rOutPerp.plus(rOutParallel));
    }

    /**
     * Rotates a vector around an {@code axis}.
     * @param vec a vector
     * @param axis the rotation axis
     * @param angle the angle in radians
     * @return the rotated vector
     */
    public static @NotNull Vec3 rotate(@NotNull Vec3 vec, @NotNull Vec3 axis, double angle) {
        Vec3 vxp = axis.cross(vec);
        Vec3 vxvxp = axis.cross(vxp);
        return vec.plus(vxp.times(Math.sin(angle))).plus(vxvxp.times(1 - Math.cos(angle)));
    }

    /**
     * {@return the euclidean distance between two vectors}
     * @param a a vector
     * @param b another vector
     */
    public static double distance(@NotNull Vec3 a, @NotNull Vec3 b) {
        return a.minus(b).length();
    }

    /**
     * Computes a running average of vectors.
     * @param current the current running average
     * @param next the next vector
     * @param index the one-based index of the next vector
     * @return the new running average
     */
    public static @NotNull Vec3 average(@NotNull Vec3 current, @NotNull Vec3 next, int index) {
        var factor = 1d / index;
        return new Vec3(
                Math.fma(factor, next.x() - current.x(), current.x()),
                Math.fma(factor, next.y() - current.y(), current.y()),
                Math.fma(factor, next.z() - current.z(), current.z())
        );
    }

    /**
     * {@return a component-wise maximum vector}
     * @param a a vector
     * @param b another vector
     */
    public static @NotNull Vec3 max(@NotNull Vec3 a, @NotNull Vec3 b) {
        return new Vec3(
                Math.max(a.x(), b.x()),
                Math.max(a.y(), b.y()),
                Math.max(a.z(), b.z())
        );
    }

    /**
     * {@return a component-wise minimum vector}
     * @param a a vector
     * @param b another vector
     */
    public static @NotNull Vec3 min(@NotNull Vec3 a, @NotNull Vec3 b) {
        return new Vec3(
                Math.min(a.x(), b.x()),
                Math.min(a.y(), b.y()),
                Math.min(a.z(), b.z())
        );
    }

    /**
     * {@return <code>a * b + c</code>}
     * @param a scalar
     * @param b a vector
     * @param c another vector
     */
    public static @NotNull Vec3 fma(double a, @NotNull Vec3 b, @NotNull Vec3 c) {
        return new Vec3(
                Math.fma(a, b.x(), c.x()),
                Math.fma(a, b.y(), c.y()),
                Math.fma(a, b.z(), c.z())
        );
    }

    public static double tripleProduct(@NotNull Vec3 a, @NotNull Vec3 b, @NotNull Vec3 c) {
        return a.x * b.y * c.z + a.y * b.z * c.x + a.z * b.x * c.y - c.x * b.y * a.z - c.y * b.z * a.x - c.z * b.x * a.y;
    }

    public @NotNull Vec3 plus(double x, double y, double z) {
        return new Vec3(this.x + x, this.y + y, this.z + z);
    }

    public @NotNull Vec3 minus(double x, double y, double z) {
        return new Vec3(this.x - x, this.y - y, this.z - z);
    }

    /**
     * Adds a vector to this vector
     * @param other a vector
     * @return the sum of this and the other vector
     */
    public @NotNull Vec3 plus(@NotNull Vec3 other) {
        return new Vec3(this.x + other.x, this.y + other.y, this.z + other.z);
    }

    /**
     * Subtracts a vector from this vector
     * @param other a vector
     * @return the difference of this and the other vector
     */
    public @NotNull Vec3 minus(@NotNull Vec3 other) {
        return new Vec3(this.x - other.x, this.y - other.y, this.z - other.z);
    }

    /**
     * Computes the scalar product of this and another vector
     * @param other a vector
     * @return the scalar product
     */
    public double times(@NotNull Vec3 other) {
        return this.x * other.x + this.y * other.y + this.z * other.z;
    }

    /**
     * Multiplies this vector with a scalar
     * @param t a scalar
     * @return the product of this vector and the scalar
     */
    public @NotNull Vec3 times(double t) {
        return new Vec3(this.x * t, this.y * t, this.z * t);
    }

    /**
     * Negates this vector.
     * {@return the negated vector}
     */
    public @NotNull Vec3 neg() {
        return new Vec3(-x, -y, -z);
    }

    /**
     * Inverts each component of this vector.
     * @return the inverted vector.
     */
    public @NotNull Vec3 inv() {
        return new Vec3(1 / x, 1 / y, 1 / z);
    }

    /**
     * Computes the cross-product of this and another vector
     * @param other a vector
     * @return the cross-product
     */
    public @NotNull Vec3 cross(@NotNull Vec3 other) {
        return new Vec3(
                Math.fma(this.y, other.z, - other.y * this.z),
                Math.fma(this.z, other.x, - other.z * this.x),
                Math.fma(this.x, other.y, - other.x * this.y)
        );
    }

    /**
     * Divides this vector by a scalar
     * @param t a scalar
     * @return this vector divided by the scalar
     */
    public @NotNull Vec3 div(double t) {
        return new Vec3(this.x / t, this.y / t, this.z / t);
    }

    /**
     * {@return the squared length of this vector}
     */
    public double squared() {
        return this.x * this.x + this.y * this.y + this.z * this.z;
    }

    /**
     * {@return the length of this vector}
     */
    public double length() {
        return Math.sqrt(squared());
    }

    /**
     * {@return whether this vector is near zero}
     */
    public boolean isNearZero() {
        var s = 1e-8;
        return Math.abs(x) < s && Math.abs(y) < s && Math.abs(z) < s;
    }

    /**
     * {@return a unit vector with the same direction as this vector}
     */
    public @NotNull Vec3 unit() {
        var squared = squared();
        if (squared == 1) return this;
        return div(Math.sqrt(squared));
    }

    /**
     * {@return the n-th component of this vector}
     * @param axis the component index
     */
    public double get(int axis) {
        return switch (axis) {
            case 0 -> x;
            case 1 -> y;
            case 2 -> z;
            default -> throw new IndexOutOfBoundsException(axis);
        };
    }

    public @NotNull Vec3 withX(double x) {
        return new Vec3(x, y, z);
    }

    public @NotNull Vec3 withY(double y) {
        return new Vec3(x, y, z);
    }

    public @NotNull Vec3 withZ(double z) {
        return new Vec3(x, y, z);
    }

    @Override
    public @NotNull String toString() {
        return "(" + x + "," + y + "," + z + ")";
    }
}
